from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn import metrics
import numpy as np
from scipy import sparse
import logging

def cvmRun(dict_conf, M, colnames, rownames):
    
    '''
    Run a k-fold random forest. 
    :param dict_conf: Dictionary containing setting parameters. 
    :param M: Matrix containing features. 
    :param colnames: Array of columns names. 
    :param rownames: Array of rows names. 
    :return: Predictions
    '''

    np.random.seed(1102)

    M = sparse.csc_matrix(M)

    y = np.array(M[:, colnames.index[colnames == dict_conf["target"]]].todense()).flatten()
    if dict_conf["target_type"] == "classification":
        y = np.array((y > dict_conf["threshold"]) + 0)
    if sum(y == 0) == len(y):
            logging.warning("Target " + dict_conf["target"] + " | Target always zero!")
            y[0] = 1
    
    X = M[:, colnames.index[colnames.isin(dict_conf["features"])]]
    
    random_folds = np.random.choice(range(dict_conf["nfolds"]), size=len(y)) + 1
    while len(np.unique(random_folds)) != dict_conf["nfolds"]:
        random_folds = np.random.choice(range(dict_conf["nfolds"]), size=len(y)) + 1
    
    predictions = np.zeros(len(y))

    if dict_conf["target_type"] == "classification":
        model = RandomForestClassifier(
            n_estimators=dict_conf["n_estimators"], 
            max_depth=dict_conf["max_depth"], 
            random_state=1102
        )
    else:
        model = RandomForestRegressor(
            n_estimators=dict_conf["n_estimators"], 
            max_depth=dict_conf["max_depth"], 
            random_state=1102
        ) 

    for k in range(dict_conf["nfolds"]):
        k = k + 1
        logging.debug("Target " + dict_conf["target"] + " | Fold " + str(k) + ".")
        bln_tmp = (random_folds != k)
        X_tmp = X[rownames.index[bln_tmp], ]
        y_tmp = y[bln_tmp]
        if sum(y_tmp == 0) == len(y_tmp):
            logging.warning("Target " + dict_conf["target"] + " | Fold " + str(k) + " | Target always zero!.")
            y_tmp[0] = 1
        model.fit(X_tmp, y_tmp)
        if dict_conf["target_type"] == "classification":
            predictions_tmp = model.predict_proba(X[rownames.index[~bln_tmp], :])
            predictions[~bln_tmp] = predictions_tmp[:, model.classes_ == 1].flatten()
        else:
            predictions_tmp = model.predict(X[rownames.index[~bln_tmp], :])
            predictions[~bln_tmp] = predictions_tmp

    if dict_conf["target_type"] == "classification":
        fpr, tpr, _ = metrics.roc_curve(y_true=y+1, y_score=predictions, pos_label=2)
        logging.debug("Target " + dict_conf["target"] + " | AUC on training set: " + str(metrics.auc(fpr, tpr)) + ".")
    else: 
        logging.debug("Target " + dict_conf["target"] + " | RMSE on training set: " + str( np.mean( (predictions-y)**2 )**(0.5) ) + ".")

    return predictions



import multiprocessing
import functools

def cvmMultiRun(array_dict_conf, M, colnames, rownames, npool=6):
    '''
    Run cvmRun in multiprocessing. 
    :param array_dict_conf: List of dictionaries containing setting parameters. 
    :param M: Matrix containing features. 
    :param colnames: Array of columns names. 
    :param rownames: Array of rows names.
    :param npool: Number of workers.  
    :return: Predictions
    '''
    pool = multiprocessing.Pool(npool)
    output = pool.map(functools.partial(cvmRun, M=M, colnames=colnames, rownames=rownames), array_dict_conf)
    return output




from scipy import sparse
from sklearn.preprocessing import normalize
import numpy as np
import logging
from .cosimtop import cosimtop

class CFM:

    '''
    Preferences Matrix Class
    '''
    

    def __init__(self, M):

        '''
        Initialization. 
        :param M: Preferences matrix, each row represents a user, each column represents the product.  
        '''
        
        # Computations are performed in sparse matrices in order to hold more data. 
        self.M = sparse.csr_matrix(M)
        logging.debug("M matrix has shape " + str(self.M.shape[0]) + "x" + str(self.M.shape[1]) + ".")


    def computeSimilarityMatrix(self, bln_bin = False, bln_norm = True, flt_ths = 0.0, ntop = None, flt_lb = -1):

        '''
        Compute the cosine similarity matrix.
        :param bln_bin: Binarize each matrix entry? If True, each entry will be binarized according to the flt_ths. 
        :param bln_norm: Should the output matrix entries be constrained in the interval [-1, +1]? 
        :param flt_ths: Threshold used by the binarization. 
        :param ntop: How many most similar users to keep for each user? If None, every similarity coefficient will be kept. Set this parameter to take advantage of the sparse matrix memory optimization. 
        :param flt_lb: Lower bound value for the output similarity matrix entries. 
        '''

        if ntop is None:
            ntop = self.M.shape[0]

        self.bln_bin = bln_bin
        self.bln_norm = bln_norm
        self.flt_ths = flt_ths
        self.ntop = ntop

        self.B = self.M.copy()
        self.B.data = (self.B.data > flt_ths) + 0.0
        logging.debug("B matrix has shape " + str(self.B.shape[0]) + "x" + str(self.B.shape[1]) + ".")
        logging.debug("B matrix has sparsity " + str(sum(self.B.data)/(self.B.shape[0] * self.B.shape[1])*100) + "%.")

        if bln_bin:
            self.S = self.B.copy()
        else:
            self.S = self.M.copy()

        if bln_norm:
            self.S = normalize(self.S, norm='l2', axis=1)

        logging.debug("Computing the similarity matrix S ...")
        self.S = cosimtop(self.S, self.S.transpose(), ntop = ntop, lower_bound=flt_lb)
        logging.debug("S matrix has shape " + str(self.S.shape[0]) + "x" + str(self.S.shape[1]) + ".")
    

    def computeScores(self):

        '''
        Compute the scores for each couple (user, product). 
        '''

        logging.debug("Computing the matrix SNORMALIZED ...")
        self.SNORMALIZED = self.S.copy()
        self.SNORMALIZED = self.SNORMALIZED - sparse.diags(self.SNORMALIZED.diagonal(), format="csr")
        self.SNORMALIZED = normalize(self.SNORMALIZED, norm='l1', axis=1)
        logging.debug("Computing the matrix SCORES ...")
        self.SCORES = self.SNORMALIZED * self.B
        logging.debug("SCORES matrix has shape " + str(self.SCORES.shape[0]) + "x" + str(self.SCORES.shape[1]) + ".")


    def computeAmounts(self):

        '''
        Compute the averaged amount for each couple (user, product).
        '''

        logging.debug("Computing the matrix AMOUNTS ...")
        self.AMOUNTS = self.SNORMALIZED * self.M
        logging.debug("AMOUNTS matrix has shape " + str(self.AMOUNTS.shape[0]) + "x" + str(self.AMOUNTS.shape[1]) + ".")
    

    def computePerformances(self):

        '''
        Compute model performances. 
        '''

        # TODO: add out of the sample testing. 

        logging.debug("Computing the performances ...")

        self.avg_global_scores_diff = abs(self.SCORES - self.B).mean()
        self.avg_pos_scores_diff = abs(self.SCORES[self.B > 0.01] - 1).mean()
        logging.debug("Average global scores difference: " + str(self.avg_global_scores_diff) + ".")
        logging.debug("Average positive scores difference: " + str(self.avg_pos_scores_diff) + ".")

        if self.avg_pos_scores_diff >= 0.5:
            logging.warning("The scores model could not perform better than a random model! Further tests needed!")

        self.avg_global_amounts_diff = abs(self.AMOUNTS - self.M).mean()
        tmp_positive_amounts = self.AMOUNTS[self.B > 0.01]
        tmp_real_positive_amounts = self.M[self.B > 0.01]
        tmp_random_model_pred = tmp_positive_amounts.mean()
        self.avg_pos_amounts_diff_perc = (abs(tmp_positive_amounts - tmp_real_positive_amounts) / tmp_real_positive_amounts).mean() * 100
        self.avg_pos_ampunts_diff_perc_random = (abs(tmp_random_model_pred - tmp_real_positive_amounts) / tmp_real_positive_amounts).mean() * 100
        logging.debug("Average global amounts difference: " + str(self.avg_global_amounts_diff) + ".")
        logging.debug("Average positive amounts difference percentage: " + str(self.avg_pos_amounts_diff_perc) + "%.")
        logging.debug("Average positive amounts difference percentage of a random model: " + str(self.avg_pos_ampunts_diff_perc_random) + "%.")

        if self.avg_pos_amounts_diff_perc >= self.avg_pos_ampunts_diff_perc_random:
            logging.warning("The amounts model could not perform better than a random model! Further tests needed!")

    
    def computeEverything(self, bln_bin = False, bln_norm = True, flt_ths = 0.0, ntop = None, flt_lb = -1):

        '''
        Method for lazy people like me. :)
        :param bln_bin: Binarize each matrix entry? If True, each entry will be binarized according to the flt_ths. 
        :param bln_norm: Should the output matrix entries be constrained in the interval [-1, +1]? 
        :param flt_ths: Threshold used by the binarization. 
        :param ntop: How many most similar users to keep for each user? If None, every similarity coefficient will be kept. Set this parameter to take advantage of the sparse matrix memory optimization. 
        :param flt_lb: Lower bound value for the output similarity matrix entries. 
        '''

        self.computeSimilarityMatrix(bln_bin = bln_bin, bln_norm = bln_norm, flt_ths = flt_ths, ntop = ntop, flt_lb = flt_lb)
        self.computeScores()
        self.computeAmounts()
        self.computePerformances()



from scipy.sparse import csr_matrix
import sparse_dot_topn.sparse_dot_topn as ct
import numpy as np

def cosimtop(A, B, ntop, lower_bound=0):

    '''
    Optimized cosine similarity computation.
    :param A: First matrix.
    :param B: Second matrix.
    :param ntop: Top n for each row.
    :param lower_bound: Lower bound for each row.
    :return: Cosine similarity matrix.
    '''
    
    A = A.tocsr()
    B = B.tocsr()
    M, _ = A.shape
    _, N = B.shape
    idx_dtype = np.int32
    nnz_max = M * ntop
    indptr = np.zeros(M + 1, dtype = idx_dtype)
    indices = np.zeros(nnz_max, dtype = idx_dtype)
    data = np.zeros(nnz_max, dtype = A.dtype)
    ct.sparse_dot_topn(
        M, N, np.asarray(A.indptr, dtype = idx_dtype),
        np.asarray(A.indices, dtype = idx_dtype),
        A.data,
        np.asarray(B.indptr, dtype = idx_dtype),
        np.asarray(B.indices, dtype = idx_dtype),
        B.data,
        ntop,
        lower_bound,
        indptr, 
        indices, 
        data
    )
    return csr_matrix(
        (data, indices, indptr), 
        shape=(M, N)
    )


from setuptools import setup

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='recoengi',
    url='https://github.com/ngshya/recoengi',
    author='ngshya',
    author_email='ngshya@gmail.com',
    packages=['recoengi', 'recoengi.cf', 'recoengi.cv', 'recoengi.sampledata'],
    install_requires=required,
    version='1.1.0',
    license='proprietary',
    description='',
    long_description=open('README.md').read(),
    package_data={
        'recoengi': ['*'],
        'recoengi': ['sampledata/*'],
        'notebooks': ['*'],
        'static': ['*'],
        'scripts': ['*']
    }
)
